G           =  6.673840e-11  # m3 / (kg s2)  # Gravitational constant
L_sun       =  3.846000e+26  # W             # Solar luminosity
M_earth     =  5.974200e+24  # kg            # Earth mass
M_jup       =  1.898700e+27  # kg            # Jupiter mass
M_sun       =  1.989100e+30  # kg            # Solar mass
N_A         =  6.022141e+23  # 1 / mol       # Avogadro's number
R           =  8.314462e+00  # J / (K mol)   # Gas constant
R_earth     =  6.378136e+06  # m             # Earth equatorial radius
R_jup       =  7.149200e+07  # m             # Jupiter equatorial radius
R_sun       =  6.955080e+08  # m             # Solar radius
Ryd         =  1.097373e+07  # 1 / m         # Rydberg constant
a0          =  5.291772e-11  # m             # Bohr radius
alpha       =  7.297353e-03  #               # Fine-structure constant
atmosphere  =  1.013250e+05  # Pa            # Atmosphere
au          =  1.495979e+11  # m             # Astronomical Unit
b_wien      =  2.897772e-03  # K m           # Wien wavelength displacement law constant
c           =  2.997925e+08  # m / s         # Speed of light in vacuum
e           =  1.602177e-19  # C             # Electron charge
eps0        =  8.854188e-12  # F / m         # Electric constant
g0          =  9.806650e+00  # m / s2        # Standard acceleration of gravity
h           =  6.626070e-34  # J s           # Planck constant
hbar        =  1.054572e-34  # J s           # Reduced Planck constant
k_B         =  1.380649e-23  # J / K         # Boltzmann constant
kpc         =  3.085678e+19  # m             # Kiloparsec
m_e         =  9.109383e-31  # kg            # Electron mass
m_n         =  1.674927e-27  # kg            # Neutron mass
m_p         =  1.672622e-27  # kg            # Proton mass
mu0         =  1.256637e-06  # N / A2        # Magnetic constant
muB         =  9.274010e-24  # J / T         # Bohr magneton
pc          =  3.085678e+16  # m             # Parsec
sigma_sb    =  5.670373e-08  # W / (K4 m2)   # Stefan-Boltzmann constant
u           =  1.660539e-27  # kg            # Atomic mass

